import app from "firebase/app";
import "firebase/auth";
import "firebase/database";
import "firebase/functions";
import "firebase/storage";
import "firebase/firestore";

const config = {
  apiKey: process.env.REACT_APP_API_KEY,
  authDomain: process.env.REACT_APP_AUTH_DOMAIN,
  databaseURL: process.env.REACT_APP_DATABASE_URL,
  projectId: process.env.REACT_APP_PROJECT_ID,
  storageBucket: process.env.REACT_APP_STORAGE_BUCKET,
  messagingSenderId: process.env.REACT_APP_MESSAGING_SENDER_ID,
  appId: process.env.REACT_APP_APP_ID
};

class Firebase {
  constructor() {
    app.initializeApp(config);

    this.auth = app.auth();
    this.db = app.database();
    this.cfs = app.firestore();
    this.functions = app.functions();
    this.storage = app.storage();
  }

  // *** Merge Auth and DB User API *** //

  onAuthUserListener = (next, fallback) =>
    this.auth.onAuthStateChanged(authUser => {
      if (authUser) {
        this.user(authUser.uid)
          .once("value")
          .then(snapshot => {
            const dbUser = snapshot.val();

            // default empty roles
            if (!dbUser.roles) {
              dbUser.roles = {};
            }

            // merge auth and db user
            authUser = {
              uid: authUser.uid,
              email: authUser.email,
              ...dbUser
            };

            next(authUser);
          });
      } else {
        fallback();
      }
    });

  // *** Auth API ***

  doCreateUserWithEmailAndPassword = (email, password) =>
    this.auth.createUserWithEmailAndPassword(email, password);

  doSignInWithEmailAndPassword = (email, password) =>
    this.auth.signInWithEmailAndPassword(email, password);

  doSignOut = () => this.auth.signOut();

  doPasswordReset = email => this.auth.sendPasswordResetEmail(email);

  doPasswordUpdate = password => this.auth.currentUser.updatePassword(password);

  // *** User API ***

  users = () => this.db.ref("users");

  user = uid => this.db.ref(`users/${uid}`);

  userPayments = uid => this.cfs.collection(`users/${uid}/payments`);

  userToken = uid => this.auth.currentUser.getIdToken(/* forceRefresh */ true);

  getCellsFromSquareGeoRegion = () =>
    this.functions.httpsCallable("getCellsFromSquareGeoRegion");

  // *** Lots API ***

  lots = () => this.db.ref("lots");

  lotByUid = uid => {
    return this.functions.httpsCallable("lotByUid")(uid);
  };
  lotByCoords = coords => {
    return this.functions.httpsCallable("getLot")(coords);
  };
  lotsByCoordsFast = coords => {
    return this.functions.httpsCallable("getLotsCluster")(coords);
  };
  lotsByCoords = coords => {
    return this.functions.httpsCallable("getLots")(coords);
  };

  lotsByUser = () => this.functions.httpsCallable("userlots")();

  buyLot = lotid => this.functions.httpsCallable("buy")(lotid);

  rentLot = ({ lotid, period }) =>
    this.functions.httpsCallable("rent")({ lotid, period });

  setLotStatus = ({ lotid, status, userPrice = 0, rentPrice = 0 }) =>
    this.functions.httpsCallable("changeStatus")({
      lotid,
      status,
      userPrice,
      rentPrice
    });

  cellsByCoord = coords => this.getCellsFromSquareGeoRegion()(coords);

  lotsWithStructures = () => this.db.ref("lots").orderByKey();

  // *** Storage API *** //

  getStorageList = path => {
    const storageRef = this.storage.ref();
    const listRef = storageRef.child(path);
    return listRef.listAll();
  };

  getStructureScreenshots = str_uid => {
    const storageRef = this.storage.ref();
    // Create a reference under which you want to list
    const listRef = storageRef.child(`screenshots/${str_uid}`);

    // Find all the prefixes and items.
    listRef
      .listAll()
      .then(function(res) {
        res.prefixes.forEach(function(folderRef) {
          console.log(folderRef);
          // All the prefixes under listRef.
          // You may call listAll() recursively on them.
        });
        res.items.forEach(function(itemRef) {
          console.log(itemRef.getDownloadURL());
          // All the items under listRef.
        });
      })
      .catch(function(error) {
        console.error(error);
        // Uh-oh, an error occurred!
      });
  };
}

export default Firebase;
