import React, { Fragment } from "react";
import { connect } from "react-redux";
import { compose } from "recompose";
import { withFirebase } from "../Firebase";
import { withRouter } from "react-router-dom";

import MapBase from "./MapBase";
import LotPopup from "./LotPopup";
// import StatusFilters from "../StatusFilters/StatusFilters";
import "./Map.scss";
import mapboxgl from "mapbox-gl";

import { STATUS, FILTER } from "../../constants/lot";
import * as ROUTES from "../../constants/routes";

import { updateAuthUserBalance } from "../../actions/user";
import { requestListRefresh } from "../../actions/lot";

import * as BIT from "../../utils/Bitwise";
import { async } from "@firebase/util";

const FILL_OPACITY = 0.2;

const LINE_WIDTH = 1;

const LOT_PADDING_K = 0.03;
const LOT_PADDING_K2 = 0.67;

const GP_ACCURACY_THRESHOLD = 1000;

const colors = {
  blue: "#007AFF",
  green: "#4AAA48",
  grey: "#979797",
  transparent: "rgba(0, 0, 0, 0)"
};
const ownColors = {
  fill: colors.blue,
  stroke: colors.blue
};
const anotherColors = {
  fill: colors.green,
  stroke: colors.green
};
const restColors = {
  fill: colors.transparent,
  stroke: colors.grey
};
const dict = {
  own: { ...ownColors },
  ownForSale: { ...ownColors },
  ownForRent: { ...ownColors },
  ownRented: { ...ownColors },
  ownLeased: { ...ownColors },
  ownAuction: { ...ownColors },
  forSale: { ...anotherColors },
  forSaleOwned: { ...anotherColors },
  forRent: { ...anotherColors },
  auction: { ...anotherColors },
  rest: { ...restColors }
};

const CELL_SIZE = 6.77;

const EMPTY_COLLECTION = {
  type: "FeatureCollection",
  features: []
};

class Map extends MapBase {
  constructor(props) {
    super(props);

    const lastPosition = JSON.parse(
      localStorage.getItem("avvy:lastPosition") || "{}"
    );

    this.state = {
      ...this.state,
      lng: lastPosition.lng || 37.635308,
      lat: lastPosition.lat || 55.737694,
      zoom: 16,
      selectedLots: false,
      clusterPos: [],
      cells: {}
    };

    this.lots = {};
    for (let entry in dict) {
      this.lots[entry] = {
        type: "FeatureCollection",
        features: []
      };
    }

    this.popup = null;
  }

  componentDidUpdate(prevProps, prevState) {
    const { filter } = this.props;
    if (filter !== prevProps.filter) this.drawCells();
  }

  calcScale() {
    const { zoom } = this.state;
    return zoom >= 16 ? 1 : Math.pow(2, Math.ceil(17 - zoom));
  }

  drawCells() {
    this.closePopup();
    const { cells } = this.state;
    const { user, filter } = this.props;
    const { zoom } = this.state;
    const paddingLeft = zoom < 16 ? LOT_PADDING_K2 : LOT_PADDING_K;
    const paddingRight = zoom < 16 ? LOT_PADDING_K2 : LOT_PADDING_K;

    for (let fc in this.lots) this.lots[fc].features = [];

    for (let key in cells) {
      const cell = cells[key];

      const [p1, p2, p3, p4] = cell.coordinates;

      var offsetX = (p2._longitude - p1._longitude) * paddingRight;
      var offxetY = (p1._latitude - p3._latitude) * paddingRight;

      var offsetXS = (p2._longitude - p1._longitude) * paddingLeft;
      var offxetYS = (p1._latitude - p3._latitude) * paddingLeft;

      var poly = {
        type: "Feature",
        properties: {
          position: [
            p1._latitude + (p4._latitude - p1._latitude) / 2,
            p1._longitude + (p3._longitude - p1._longitude) / 2
          ]
        },
        geometry: {
          type: "Polygon",
          coordinates: [[]]
        }
      };
      poly.geometry.coordinates[0].push(
        [p1._longitude + offsetXS, p1._latitude - offxetYS],
        [p2._longitude - offsetX, p2._latitude - offxetYS],
        [p3._longitude - offsetX, p3._latitude + offxetY],
        [p4._longitude + offsetXS, p4._latitude + offxetY]
      );
      poly.geometry.coordinates[0].push(poly.geometry.coordinates[0][0]);

      const own = cell.owner === user.uid;
      switch (cell.status) {
        case STATUS.FOR_SALE:
          this.lots[
            own ? "ownForSale" : cell.owner ? "forSaleOwned" : "forSale"
          ].features.push(poly);
          break;
        case STATUS.FOR_RENT:
          this.lots[own ? "ownForRent" : "forRent"].features.push(poly);
          break;
        case STATUS.RENTED:
          this.lots[
            own ? "ownRented" : cell.renter === user.uid ? "ownLeased" : "rest"
          ].features.push(poly);
          break;
        default:
          this.lots[own ? "own" : "rest"].features.push(poly);
          break;
      }
    }

    for (let entry in dict) {
      if (
        entry === "own" ||
        entry === "rest" ||
        (entry === "ownForSale" &&
          BIT.isset(filter, FILTER.FOR_SALE + FILTER.OWN_SHIFT)) ||
        (entry === "ownForRent" &&
          BIT.isset(filter, FILTER.FOR_RENT + FILTER.OWN_SHIFT)) ||
        (entry === "ownRented" &&
          BIT.isset(filter, FILTER.RENTED + FILTER.OWN_SHIFT)) ||
        (entry === "ownLeased" && BIT.isset(filter, FILTER.LEASED)) ||
        (entry === "ownAuction" &&
          BIT.isset(filter, FILTER.AUCTION + FILTER.OWN_SHIFT)) ||
        (entry === "forSale" && BIT.isset(filter, FILTER.FREE_SALE)) ||
        (entry === "forSaleOwned" && BIT.isset(filter, FILTER.FOR_SALE)) ||
        (entry === "forRent" && BIT.isset(filter, FILTER.FOR_RENT)) ||
        (entry === "auction" && BIT.isset(filter, FILTER.AUCTION))
      ) {
        this.map.getSource(`${entry}-border`).setData(this.lots[entry]);
        this.map.getSource(`${entry}-fill`).setData(this.lots[entry]);
      } else {
        this.map.getSource(`${entry}-border`).setData(EMPTY_COLLECTION);
        this.map.getSource(`${entry}-fill`).setData(EMPTY_COLLECTION);
      }
    }
  }

  updateCellsInView(callback) {
    const { firebase } = this.props;
    const { lng, lat } = this.map.getCenter();
    const zoom = this.map.getZoom().toFixed(2);

    this.setState({
      lng: lng,
      lat: lat,
      zoom: zoom,
      loading: true
    });

    let scale = this.calcScale();

    if (scale > 1) {
      firebase
        .cellsByCoord({
          longitude: lng,
          latitude: lat,
          width: 120 * scale, //120
          height: 70 * scale, //70
          scale: scale
        })
        .then(data => {
          const cells = JSON.parse(data.data.lots);
          this.setState({ cells });
          this.drawCells();
        })
        .finally(() => {
          this.setState({ loading: false });
        });
    } else {
      firebase
        .lotsByCoordsFast({
          longitude: lng,
          latitude: lat,
          width: 200 * scale, //120
          height: 200 * scale, //70
          scale: scale
        })
        .then(data => {
          const cells = JSON.parse(data.data.lots);
          this.setState({ cells });
          this.drawCells();
        })
        .finally(() => {
          this.setState({ loading: false });
        });
    }
  }

  onMapLoad() {
    var context = this;

    this.setState({ selectedLots: false });

    var content = document.getElementById("lot-popup-content");
    this.popup = new mapboxgl.Popup({ closeOnClick: true, closeButton: true })
      .setDOMContent(content)
      .setLngLat([0, 0])
      .addTo(this.map);

    this.popup.remove = function() {
      context.popup._container.classList.add("hide");
    };
    this.popup.remove();

    const addLayer = (id, data, fillcolor, linecolor) => {
      const config = {
        id: id,
        source: {
          type: "geojson",
          data: data
        }
      };
      const fillpaint = {
        paint: {
          "fill-color": fillcolor,
          "fill-opacity": FILL_OPACITY
        }
      };
      const linepaint = {
        layout: {
          "line-join": "round",
          "line-cap": "round"
        },
        paint: {
          "line-color": linecolor || fillcolor,
          "line-width": LINE_WIDTH
        }
      };
      this.map.addLayer({
        ...config,
        id: `${id}-border`,
        type: "line",
        ...linepaint
      });
      this.map.addLayer({
        ...config,
        id: `${id}-fill`,
        type: "fill",
        ...fillpaint
      });
    };
    for (let entry in dict) {
      const colors = dict[entry];
      addLayer(entry, this.lots[entry], colors.fill, colors.stroke);
      this.map.on("dblclick", `${entry}-fill`, this.onMapDblclick.bind(this));
    }

    this.map.on("zoom", this.onMapZoom.bind(this));
    this.map.on("click", `forSale-fill`, this.onMapClick.bind(this));
    this.map.on("click", `forSaleOwned-fill`, this.onMapClick.bind(this));
    this.map.on("click", `forRent-fill`, this.onMapClick.bind(this));

    if (navigator.geolocation) {
      navigator.geolocation.getCurrentPosition(
        position => {
          const { longitude, latitude, accuracy } = position.coords;
          const pos = { lng: longitude, lat: latitude };
          if (accuracy <= GP_ACCURACY_THRESHOLD) {
            const rfn = function(cnt = 0) {
              if (this.state.loading) {
                setTimeout(() => {
                  rfn.call(this, cnt++);
                }, 100);
              } else {
                localStorage.setItem("avvy:lastPosition", JSON.stringify(pos));
                this.map.setCenter(pos);
              }
            };
            rfn.call(this);
          }
        },
        positionError => {
          console.log("POSITION ERROR", positionError);
        }
      );
    }
    this.updateCellsInView();
  }

  onMapZoom() {
    this.closePopup();
  }

  onMapClick(e) {
    if (this.state.loading) return;

    const { firebase } = this.props;

    const pos = JSON.parse(e.features[0].properties.position);

    let scale = this.calcScale();

    this.setState({
      selectedLots: false,
      clusterPos: pos
    });

    this.showPopup();

    if (scale > 1) {
      firebase
        .lotsByCoords({
          longitude: pos[1],
          latitude: pos[0],
          width: CELL_SIZE * scale, //120
          height: CELL_SIZE * scale //70
        })
        .then(cells => {
          this.setState({ selectedLots: JSON.parse(cells.data.lots) });
        });
    } else {
      firebase
        .lotByCoords({
          longitude: pos[1],
          latitude: pos[0]
        })
        .then(cells => {
          this.setState({ selectedLots: JSON.parse(cells.data.lot) });
        });
    }
  }

  onMapDblclick(e) {
    if (this.state.loading) return;

    e.preventDefault();
    this.closePopup();
    let scale = this.calcScale();
    if (scale === 1) {
      const { firebase } = this.props;
      const pos = JSON.parse(e.features[0].properties.position);
      firebase
        .lotByCoords({
          longitude: pos[1],
          latitude: pos[0]
        })
        .then(cells => {
          const id = Object.keys(JSON.parse(cells.data.lot))[0];
          this.props.history.push(ROUTES.LOT.replace(/:[^:]+$/, id));
        });
    }
  }

  onMapMoveEnd() {
    const { lng, lat, zoom } = this.state;

    console.log("onMapMoveEnd", lng, lat, zoom, arguments);

    this.setState({
      lng: lng,
      lat: lat,
      zoom: zoom
    });
    localStorage.setItem(
      "avvy:lastPosition",
      JSON.stringify({
        lng: lng,
        lat: lat
      })
    );

    this.updateCellsInView();
  }

  showPopup() {
    const { clusterPos } = this.state;

    this.popup.setLngLat([clusterPos[1], clusterPos[0]]);

    this.popup._container.classList.remove("hide");
  }

  closePopup() {
    this.popup._container.classList.add("hide");
  }

  onBuy() {
    const {
      firebase,
      updateAuthUserBalance,
      user,
      requestListRefresh
    } = this.props;
    this.closePopup();
    this.updateCellsInView();
    firebase
      .user(user.uid)
      .once("value")
      .then(snapshot => {
        const balance = snapshot.val().Balance;
        updateAuthUserBalance(balance);
        requestListRefresh(3);
      });
  }

  render() {
    const { zoom, selectedLots } = this.state;

    return (
      <Fragment>
        <LotPopup
          lots={selectedLots}
          canBuy={zoom >= 16}
          callback={this.onBuy.bind(this)}
        />
        {/* <StatusFilters /> */}
        {super.render()}
      </Fragment>
    );
  }
}

const mapStateToProps = state => {
  const { authUser } = state.session;
  const { filter } = state.lot;
  return { user: authUser, filter };
};
const mapDispatchToProps = dispatch => ({
  updateAuthUserBalance: value => dispatch(updateAuthUserBalance(value)),
  requestListRefresh: value => dispatch(requestListRefresh(value))
});

export default compose(
  withRouter,
  withFirebase,
  connect(
    mapStateToProps,
    mapDispatchToProps
  )
)(Map);
